import inspect
from functools import wraps
import threading


class Manager(object):

    def __init__(self):
        self._registered = {}
        self._singletons = {}
        self._threadlocals = threading.local()

    def register(self, func, singleton=False, threadlocal=False):
        """
        Register a dependency function
        """
        func._giveme_singleton = singleton
        func._giveme_threadlocal = threadlocal
        
        self._registered[func.__name__] = func
        return func

    def remove(self, name):
        """
        Remove a dependency by name
        """
        del self._registered[name]

    def get(self, name):
        """
        Get a dependency factory by name, None if not registered
        """
        return self._registered.get(name)

    def get_value(self, name):
        """
        Get return value of a dependency factory or
        a live singleton instance.
        """
        factory = self._registered.get(name)
        if not factory:
            raise KeyError('Name not registered')
        if factory._giveme_singleton:
            if name in self._singletons:
                return self._singletons[name]
            self._singletons[name] = factory()
            return self._singletons[name]
        elif factory._giveme_threadlocal:
            if hasattr(self._threadlocals, name):
                return getattr(self._threadlocals, name)
            setattr(self._threadlocals, name, factory())
            return getattr(self._threadlocals, name)
        return factory()

    def clear(self):
        self._registered = {}
        self._singletons = {}


manager = Manager()


def register(function=None, singleton=False, threadlocal=False):
    """
    Register a dependency factory in the dependency manager. The function name is the
    name of the dependency.
    This can be used as a decorator.

    Args:
        function (callable): The dependency factory function
            Not needed when used as decorator.
        singleton (``bool``, optional): If ``True`` the given function is only called once
            during the application lifetime. Injectees will receive the already created
            instance when available. Defaults to ``False``
        threadlocal (``bool``, optional): Same as singleton except the returned instance
            is available only to the thread that created it. Defaults to ``False``
    """
    def decorator(function):
        return manager.register(function, singleton=singleton, threadlocal=threadlocal)
    if function:
        return decorator(function)
    else:
        return decorator





def inject(func):
    """
    Inject a dependency into given function's arguments.
    Can be used as a decorator.

    Injectee should mention named dependencies as keyword arguments.

    def db_connection():
        return create_db_connection()

    def save_thing(thing, db_connection=None):
        db_connection.store(thing)

    Args:
        func (callable): The function that accepts a dependency.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        signature = inspect.signature(func)
        params = signature.parameters
        if not params:
            return func(*args, **kwargs)

        
        factories = []

        for name, param in params.items():
            if param.kind not in (param.KEYWORD_ONLY, param.POSITIONAL_OR_KEYWORD):
                continue
            if name in kwargs:
                # Manual override, ignore it
                continue
            try:
                kwargs[name] = manager.get_value(name)
            except KeyError:
                pass
        return func(*args, **kwargs)
    return wrapper



import pytest
import time
from functools import wraps
from multiprocessing.pool import ThreadPool

from giveme import register, inject


def test_inject():
    @register
    def something():
        return 124

    @inject
    def do_some(something=None):
        return something

    assert do_some() == 124


def test_with_other_params():
    @register
    def something():
        return 124

    @inject
    def do_some(a, b, something=None):
        return a, b, something

    assert do_some(2, 4) == (2, 4, 124)


def test_with_kwargs():
    @register
    def something():
        return 124

    @inject
    def do_some(a, b, something=None, c=7):
        return a, b, something, c

    assert do_some(2, 4) == (2, 4, 124, 7)
    assert do_some(2, 4, c=12) == (2, 4, 124, 12)


def test_manual_override_kwargs():
    @register
    def something():
        return 'something'

    @inject
    def do_some(a, b=2, something=None, **kwargs):
        return a, b, something, kwargs

    assert do_some(2, b=3, something='nothing', c=4) == (2, 3, 'nothing', {'c': 4})
    assert do_some(1, 2, something='nothing') == (1, 2, 'nothing', {})


def test_manual_override_args_kwargs():
    @register
    def something():
        return 'something'

    @inject
    def do_some(*args, something=None, **kwargs):
        return args, something, kwargs
    r4 = tuple(range(4))
    assert do_some(*r4) == (r4, 'something', {})
    assert do_some(*r4, a='a') == (r4, 'something', {'a': 'a'})


def test_nested():
    @register
    def something():
        return 128

    @register
    @inject
    def something_else(something=None):
        return something*2

    @inject
    def do_some(something_else=None):
        return something_else

    assert do_some() == 128*2


def test_with_outside_decorator():
    @register
    def something():
        return 128

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            kwargs['c'] = 'changed'
            return func(*args, **kwargs)
        return wrapper

    @inject
    @decorator
    def do_some(a, something, c='original'):
        return a, something, c

    @decorator
    @inject
    def do_some_flipped(a, something, c='original'):
        return a, something, c

    assert do_some(12) == (12, 128, 'changed')
    assert do_some_flipped(12) == (12, 128, 'changed')


def test_instance_method():
    @register
    def something():
        return 128

    class SomeClass(object):

        @inject
        def do_some(self, something):
            return something

    s = SomeClass()
    assert s.do_some() == 128


def test_not_singleton():
    class Something(object):
        def __init__(self):
            self.size = 22

    @register
    def something():
        return Something()

    @inject
    def do_some(something):
        something.size = 42
        return something

    @inject
    def do_some_again(something):
        return something

    assert do_some() is not do_some_again()
    assert do_some().size == 42 and do_some_again().size == 22


def test_singleton():
    class Something(object):
        def __init__(self):
            self.size = 22

    @register(singleton=True)
    def something():
        return Something()

    @inject
    def do_some(something):
        something.size = 42
        return something

    @inject
    def do_some_again(something):
        return something

    assert do_some() is do_some_again()
    assert do_some().size == 42 and do_some_again().size == 42


def test_threadlocal():
    @register(threadlocal=True)
    def something():
        return [1, 2, 3]

    @inject
    def do_some(add, something=None):
        time.sleep(0.5)  # Make sure they run in separate threads
        something.append(add)
        return something

    tp = ThreadPool()

    t1 = tp.apply_async(do_some, args=(4, ))
    t2 = tp.apply_async(do_some, args=(5, ))

    r1 = t1.get()
    r2 = t2.get()

    assert r1 == [1, 2, 3, 4]
    assert r2 == [1, 2, 3, 5]


def test_no_decorator():
    def something():
        return [1, 2, 3]

    def do_some(something):
        something.append(4)
        return something

    register(something)
    do_some = inject(do_some)

    assert do_some() == [1, 2, 3, 4]



from setuptools import setup

setup(
    name='giveme',
    version='0.3.1',
    description='Tiny dependency injection framework',
    author='Steinthor Palsson',
    author_email='steini90@gmail.com',
    url='https://github.com/steinitzu/giveme',
    license='MIT',
    install_requires=[],
    packages=['giveme']
)
